﻿using Jypeli;
using System;
using static Jypeli.Game;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Mailaolion luokka.
    /// </summary>
    class Maila : PhysicsObject
    {
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu tai animaatio on suoritettu loppuun
        /// Aloitusanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa        
        /// Lopetusanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa
        /// Tuhottu = olio on tuhoutumassa tai on jo tuhottu        
        /// </code>
        /// </remarks>        
        private enum OlionTilanne : int { Normaali, Aloitusanimaatio, Lopetusanimaatio, Tuhottu }

        /// <summary>
        /// Mailan reunan ja pelialueen reunan väliin jätettävä tila pikseleinä.
        /// </summary>
        private const int RAJAN_MARGINAALI = 5;

        /// <summary>
        /// Lopetusanimaatiossa mailan kokoa pienennetään tämän kertoimen mukaan.
        /// </summary>
        private const float LOPETUSANIMAATIO_KOON_KERROIN_PER_PAIVITYS = 0.9f;

        /// <summary>
        /// Animaatio lopetetaan kun mailan koko on pienempi.
        /// </summary>
        private const float ANIMAATIO_PIENIN_KOON_KERROIN = 0.1f;

        /// <summary>
        /// Kuinka monta pikseliä mailaa siirretään animaatiossa per päivitys.
        /// </summary>
        private const int ANIMAATIO_POSITION_MUUTOS_PER_PAIVITYS = 2;

        /// <summary>
        /// Varjokuvan siirto pikseleinä vaakasuunnassa mailaan verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_VAAKASUUNNASSA = -3;

        /// <summary>
        /// Varjokuvan siirto pikseleinä pystysuunnassa mailaan verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_PYSTYSUUNNASSA = -7;
                        
        /// <summary>
        /// Mailan sijainnin rajoitus vasemmalla (X-koordinaatti).
        /// </summary>
        private readonly int vasenRaja;

        /// <summary>
        /// Mailan sijainnin rajoitus oikealla (X-koordinaatti).
        /// </summary>
        private readonly int oikeaRaja;

        /// <summary>
        /// Mailan vakionopeus.
        /// </summary>
        private readonly int vakionopeus;

        /// <summary>
        /// Mailan varjon GameObject-olio.
        /// </summary>
        private GameObject varjoOlio;

        /// <summary>
        /// Mailaolio siirretään tähän Y-koordinaattiin alkuanimaatiossa.
        /// </summary>
        private int alkuanimaationPaatepisteY;

        /// <summary>
        /// Olion tilanne.
        /// </summary>
        /// <remarks>Normaali, Aloitusanimaatio, Lopetusanimaatio, Tuhottu</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun mailan alkuanimaatio on suoritettu loppuun.</remarks>
        public event EventHandler MailaValmis;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun mailan lopetusanimaatio on suoritettu loppuun.</remarks>
        public event EventHandler MailanVoiPoistaa;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo on osunut mailaan.</remarks>
        public event EventHandler<AaniefektienTyypit> SoitaAaniefekti;

        /// <summary>
        /// Mailan konstruktori.
        /// </summary>
        /// <param name="x">Keskipiste X-akselilla.</param>
        /// <param name="y">Keskipiste Y-akselilla.</param>
        /// <param name="leveys">Mailan leveys pikseleinä.</param>
        /// <param name="korkeus">Mailan korkeus pikseleinä.</param>
        /// <param name="vasenRaja">Pelialueen vasemman reunan sijainti.</param>
        /// <param name="oikeaRaja">Pelialueen oikean reunan sijainti.</param>
        /// <param name="vakionopeus">Mailan vakionopeus.</param>
        /// <param name="kuva">Mailan kuva.</param>
        /// <param name="varjonKuva">Mailan varjokuva.</param>
        public Maila(int x, int y, int leveys, int korkeus, int vasenRaja, int oikeaRaja, int vakionopeus, Image kuva, Image varjonKuva)  : base(leveys, korkeus)
        {
            MakeStatic(); // staattinen olio
            Restitution = 1.0; // kimmoisuus, ei muuta pallon nopeutta jos arvo on 1
            KineticFriction = 0;

            X = x;
            Y = y;
            Image = kuva;
            this.vasenRaja = vasenRaja + (int)Size.X / 2 + RAJAN_MARGINAALI;
            this.oikeaRaja = oikeaRaja - (int)Size.X / 2 - RAJAN_MARGINAALI;
            this.vakionopeus = vakionopeus;

            varjoOlio = new GameObject(Size.X, Size.Y);
            varjoOlio.Image = varjonKuva;
            varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
            varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
            Instance.Add(varjoOlio, (int)OlioidenPiirtotasot.Varjot);

            olionTilanne = OlionTilanne.Normaali;
        }


        /// <summary>
        /// Aloitetaan mailan alkuanimaatio.
        /// </summary>
        /// <param name="animaationPaatepisteY">Mailan lopullinen Y-koordinaatti.</param>
        public void AloitaAlkuanimaatio(int animaationPaatepisteY)
        {
            alkuanimaationPaatepisteY = animaationPaatepisteY;            
            Velocity = Vector.Zero;

            varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
            varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;

            olionTilanne = OlionTilanne.Aloitusanimaatio;
        }
        

        /// <summary>
        /// Liikutetaan mailaa vakionopeudella vasemmalle tai oikealle.
        /// </summary>
        /// <param name="suunta">Suuntavektori, jossa X on -1 tai 1 ja Y on 0.</param>
        public void LiikutaMailaa(Vector suunta)
        {
            if (X <= vasenRaja && suunta.X < 0)
            {
                Velocity = Vector.Zero;
            }
            else if (X >= oikeaRaja && suunta.X > 0)
            {
                Velocity = Vector.Zero;
            }
            else
            {
                Velocity = suunta * vakionopeus;
            }
        }


        /// <summary>
        /// Pallo törmäsi mailaan, lähetetään viesti, jossa pyydetään soittamaan tietty ääniefekti.
        /// </summary>
        public void PalloTormasi()
        {
            SoitaAaniefekti?.Invoke(this, AaniefektienTyypit.Maila);           
        }


        /// <summary>
        /// Aloitetaan mailan lopetusanimaatio.
        /// </summary>
        public void PoistaPelista()
        {
            Velocity = Vector.Zero;
            olionTilanne = OlionTilanne.Lopetusanimaatio;
        }


        /// <summary>
        /// Tuhotaan maila ja sen varjo.
        /// </summary>
        public override void Destroy()
        {
            olionTilanne = OlionTilanne.Tuhottu;
            varjoOlio.Destroy();           
            base.Destroy();
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin välein.
        /// Päivittää mailan animaation ja varjon sijainnin.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (!IsDestroyed)
            {
                if (olionTilanne == OlionTilanne.Aloitusanimaatio)
                {
                    Y += ANIMAATIO_POSITION_MUUTOS_PER_PAIVITYS;
                    if (Y >= alkuanimaationPaatepisteY)
                    {
                        Y = alkuanimaationPaatepisteY;
                        olionTilanne = OlionTilanne.Normaali;
                        
                        MailaValmis?.Invoke(this, EventArgs.Empty);
                    }
                }
                else if (olionTilanne == OlionTilanne.Lopetusanimaatio)
                {
                    Size = Size * LOPETUSANIMAATIO_KOON_KERROIN_PER_PAIVITYS;                    
                    varjoOlio.Size = Size;
                    Y -= ANIMAATIO_POSITION_MUUTOS_PER_PAIVITYS;

                    if (Size.X < ANIMAATIO_PIENIN_KOON_KERROIN)
                    {
                        olionTilanne = OlionTilanne.Normaali;
                                                
                        MailanVoiPoistaa?.Invoke(this, EventArgs.Empty);
                    }
                }

                varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
                varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
            }

            base.Update(time);
        }

    }

}