﻿using Jypeli;
using System;
using static Jypeli.Game;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Pallo-olion luokka.
    /// </summary>
    class Pallo : PhysicsObject
    {
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu tai animaatio on suoritettu loppuun
        /// Aloitusanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa        
        /// Lopetusanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa
        /// Tormaysanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa
        /// Tuhoutumassa = pallo ylitti pelialueen alareunan
        /// Tuhottu = olio on tuhottu        
        /// </code>
        /// </remarks>        
        private enum OlionTilanne : int { Normaali, Aloitusanimaatio, Lopetusanimaatio, Tormaysanimaatio, Tuhoutumassa, Tuhottu }

        /// <summary>
        /// Aloitusanimaatiossa pallon kokoa suurennetaan tämän kertoimen mukaan.
        /// </summary>
        private const float ALOITUSANIMAATIO_KOON_KERROIN = 1.3f;

        /// <summary>
        /// Lopetusanimaatiossa pallon kokoa pienennetään tämän kertoimen mukaan.
        /// </summary>
        private const float LOPETUSANIMAATIO_KOON_KERROIN = 0.9f;

        /// <summary>
        /// Animaatio lopetetaan kun pallon koko on pienempi.
        /// </summary>
        private const float LOPETUSANIMAATIO_PIENIN_KOON_KERROIN = 0.1f;

        /// <summary>
        /// Varjokuvan siirto pikseleinä vaakasuunnassa palloon verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_VAAKASUUNNASSA = -3;

        /// <summary>
        /// Varjokuvan siirto pikseleinä pystysuunnassa palloon verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_PYSTYSUUNNASSA = -3;

        /// <summary>
        /// Törmäysanimaation kesto millisekunteina.
        /// </summary>
        private const int TORMAYSANIMAATION_KESTO = 40;

        /// <summary>
        /// Olion tilanne.
        /// </summary>
        /// <remarks>Normaali, Aloitusanimaatio, Lopetusanimaatio, Tormaysanimaatio, Tuhoutumassa, Tuhottu</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Pallon halkaisija pikseleinä.
        /// </summary>
        private int pallonHalkaisija;
        
        /// <summary>
        /// Pallon vakionopeus.
        /// </summary>
        private int vakionopeus;

        /// <summary>
        /// Alareunan Y-koordinaatti, jonka yli siirtyessään pallo tuhoutuu.
        /// </summary>
        private int alareunaPallonTuhoutumisraja;
                   
        /// <summary>
        /// Varjon GameObject-olio.
        /// </summary>
        private GameObject varjoOlio;

        /// <summary>
        /// Törmäysefektin GameObject-olio.
        /// </summary>
        private GameObject tormaysolio;

        /// <summary>
        /// Apumuuttuja, tähän tallennetaan pallon suunta törmäyksessä.
        /// </summary>
        private Vector suuntavektori;

        /// <summary>
        /// Apumuuttuja, törmäysanimaatiossa kulunut aika.
        /// </summary>
        private int tormaysanimaatioKulunutAika = 0;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallon alkuanimaatio on suoritettu loppuun.</remarks>
        public event EventHandler PalloValmis;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallon lopetusanimaatio on suoritettu loppuun, tai pallo on tuhoutunut.</remarks>
        public event EventHandler PallonVoiPoistaa;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo on ohittanut pelialueen alareunan.</remarks>
        public event EventHandler PalloOhittiAlareunan;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo on ohittanut pelialueen alareunan.</remarks>
        public event EventHandler<AaniefektienTyypit> SoitaAaniefekti;

        /// <summary>
        /// Pallon konstruktori.
        /// </summary>
        /// <param name="x">Keskipiste X-akselilla.</param>
        /// <param name="y">Keskipiste Y-akselilla.</param>
        /// <param name="pallonHalkaisija">Pallon halkaisija pikseleinä.</param>
        /// <param name="vakionopeus">Pallon vakionopeus.</param>
        /// <param name="alareunaPallonTuhoutumisraja">Pelialueen alareunan Y-koordinaatti.</param>
        /// <param name="kuva">Pallon kuva.</param>
        /// <param name="varjokuva">Pallon varjokuva.</param>
        /// <param name="tormayskuva">Pallon törmäysanimaatiossa käytettävä kuva.</param>
        public Pallo(int x, int y, int pallonHalkaisija, int vakionopeus, int alareunaPallonTuhoutumisraja, Image kuva, Image varjokuva, Image tormayskuva) : base(pallonHalkaisija, pallonHalkaisija)
        {            
            Restitution = 1.0; // kimmoisuus, ei muuta pallon nopeutta jos arvo on 1
            KineticFriction = 0;
            CanRotate = false;

            X = x;
            Y = y;
            Image = kuva;
            //this.tormayskuva = tormayskuva;
            this.pallonHalkaisija = pallonHalkaisija;
            this.vakionopeus = vakionopeus;
            this.alareunaPallonTuhoutumisraja = alareunaPallonTuhoutumisraja;

            varjoOlio = new GameObject(pallonHalkaisija, pallonHalkaisija);
            varjoOlio.Image = varjokuva;
            varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
            varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
            Instance.Add(varjoOlio, (int)OlioidenPiirtotasot.Varjot);

            tormaysolio = new GameObject(pallonHalkaisija, pallonHalkaisija);
            tormaysolio.Image = tormayskuva;
            tormaysolio.Position = Position;
            tormaysolio.IsVisible = false;
            Instance.Add(tormaysolio, (int)OlioidenPiirtotasot.Efektit);

            olionTilanne = OlionTilanne.Normaali;            
        }


        /// <summary>
        /// Aloitetaan pallon alkuanimaatio.
        /// </summary>
        public void AloitaAlkuanimaatio()
        {                     
            Velocity = Vector.Zero;            

            varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
            varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;

            Size = new Vector(1, 1);
            varjoOlio.Size = Size;
            
            olionTilanne = OlionTilanne.Aloitusanimaatio;
        }


        /// <summary>
        /// Laukaistaan pallo annetun vektorin suuntaan.
        /// </summary>
        /// <param name="suunta">Pallon suunta.</param>
        public void LaukaisePallo(Vector suunta)
        {            
            Velocity = suunta.Normalize() * vakionopeus;
        }


        /// <summary>
        /// Pallo törmäsi johonkin olioon, aloitetaan törmäysanimaatio.
        /// </summary>
        public void PalloTormasi()
        {            
            if (olionTilanne == OlionTilanne.Normaali)
            {
                olionTilanne = OlionTilanne.Tormaysanimaatio;
                tormaysanimaatioKulunutAika = 0;
                suuntavektori = Velocity;
                Velocity = Vector.Zero;
                tormaysolio.IsVisible = true;               
            }            
        }


        /// <summary>
        /// Aloitetaan lopetusanimaatio.
        /// </summary>
        public void PoistaPelista()
        {
            Velocity = Vector.Zero;
            olionTilanne = OlionTilanne.Lopetusanimaatio;
        }


        /// <summary>
        /// Pallo on mennyt pelialueen alareunan ohitse, lähetetään viesti että pallon voi poistaa.
        /// </summary>
        public void TuhoaPallo()
        {
            Velocity = Vector.Zero;

            SoitaAaniefekti?.Invoke(this, AaniefektienTyypit.PalloTuhoutui);
            PallonVoiPoistaa?.Invoke(this, EventArgs.Empty);
        }


        /// <summary>
        /// Tuhotaan pallo ja sen varjo.
        /// </summary>
        public override void Destroy()
        {
            olionTilanne = OlionTilanne.Tuhottu;
            varjoOlio.Destroy();
            tormaysolio.Destroy();
            base.Destroy();
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin välein.
        /// Päivittää pallon animaation ja varjon sijainnin.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (!IsDestroyed)
            {
                switch (olionTilanne)
                {
                    case OlionTilanne.Normaali:
                        if (Y < alareunaPallonTuhoutumisraja)
                        {
                            // pallo meni alareunan ohitse
                            Velocity = Vector.Zero;
                            olionTilanne = OlionTilanne.Tuhoutumassa;
                            PalloOhittiAlareunan?.Invoke(this, EventArgs.Empty);
                            break;
                        }

                        // pallo on liikkeessä, pidetään pallon nopeus vakiona
                        if (Velocity.Magnitude > 0 && Velocity.Magnitude < vakionopeus) Velocity = vakionopeus * Velocity.Normalize();
                        break;

                    case OlionTilanne.Aloitusanimaatio:
                        Size = Size * ALOITUSANIMAATIO_KOON_KERROIN;
                        varjoOlio.Size = Size;                       

                        if (Size.X >= pallonHalkaisija)
                        {
                            Size = new Vector(pallonHalkaisija, pallonHalkaisija);
                            varjoOlio.Size = Size;

                            olionTilanne = OlionTilanne.Normaali;
                                                        
                            PalloValmis?.Invoke(this, EventArgs.Empty);
                        }
                        break;

                    case OlionTilanne.Lopetusanimaatio:
                        Size = Size * LOPETUSANIMAATIO_KOON_KERROIN;                       
                        varjoOlio.Size = Size;

                        if (Size.X < LOPETUSANIMAATIO_PIENIN_KOON_KERROIN)
                        {
                            olionTilanne = OlionTilanne.Normaali;
                                                       
                            PallonVoiPoistaa?.Invoke(this, EventArgs.Empty);
                        }
                        break;

                    case OlionTilanne.Tormaysanimaatio:
                        tormaysanimaatioKulunutAika += (int)time.SinceLastUpdate.TotalMilliseconds;

                        if (tormaysanimaatioKulunutAika >= TORMAYSANIMAATION_KESTO)
                        {
                            // törmäysefekti loppuu
                            tormaysolio.IsVisible = false;
                            olionTilanne = OlionTilanne.Normaali;
                            Velocity = suuntavektori;
                        }
                        break;

                    default:
                        break;
                }

                varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
                varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
                tormaysolio.Position = Position;
            }
           
            base.Update(time);
        }                        

    }

}
