﻿using Jypeli;
using static Jypeli.Game;
using System;
using System.Collections.Generic;
using System.Linq;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Alkuvalikon animoidun otsikkotekstin luokka.
    /// </summary>
    class Skrollinaytto : GameObject
    {        
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu tai animaatio on suoritettu loppuun
        /// Animaatio = olio suorittaa animaatiota              
        /// </code>
        /// </remarks>        
        private enum OlionTilanne : int { Normaali, Animaatio }

        /// <summary>
        /// Kirjaimien paikkaan tehtävä muutos per Update-funktion kutsu.
        /// </summary>
        private const double KIRJAIMIEN_PAIKAN_MUUTOS_PER_PAIVITYS = 0.03d;

        /// <summary>
        /// Kirjaimen värin vaihteluun viiveeseen tehtävä muutos.
        /// </summary>
        private const double ANIMAATION_PAIVITYSAJAN_MUUTOS = 1.0d;

        /// <summary>
        /// Kirjaimen värin vaihtelun päivitysväli.
        /// </summary>
        private const double ANIMAATION_PAIVITYSVALI = 2.0d;

        /// <summary>
        /// Kirjaimen värin vaihtelun viive.
        /// </summary>
        private const double ANIMAATION_KULUNEEN_AJAN_KERROIN = -4.0d;

        /// <summary>
        /// Kirjaimen vaakasuunnassa tapahtuvan siirron kerroin.
        /// </summary>
        private const double ANIMAATIO_MUUTOS_VAAKASUUNNASSA_KERROIN = 0.16d;

        /// <summary>
        /// Kirjaimen pystysuunnassa tapahtuvan siirron kerroin.
        /// </summary>
        private const double ANIMAATIO_MUUTOS_PYSTYSUUNNASSA_KERROIN = 10.0d;

        /// <summary>
        /// Kirjaimen animaation (värin muutos) kesto millisekunteina.
        /// Kesto valitaan satunnaisesti kahden numeron väliltä.
        /// </summary>
        private readonly int[] animaationKestonVaihteluvalit = { 400, 500 };

        /// <summary>
        /// Kirjaimien värit valitaan satunnaisesti tästä taulukosta.
        /// </summary>
        private readonly Color[] varit = { Color.Gray, Color.White, Color.Gold, Color.Blue, Color.Red };

        /// <summary>
        /// Jokaisen kirjaimen alkuperäinen piste x-akselilla.
        /// </summary>        
        private readonly int[] aloituspisteetX;

        /// <summary>
        /// Jokaisen kirjaimen alkuperäinen piste y-akselilla.
        /// </summary>      
        private readonly int[] aloituspisteetY;

        /// <summary>
        /// Jokaisen kirjaimen animaation kesto, jokaisella kirjaimella hieman eri luku.
        /// </summary>
        private readonly double[] animaationKestot;

        /// <summary>
        /// Alkuperäinen tekstin väri.
        /// </summary>
        private readonly Color tekstinVari;
              
        /// <summary>
        /// Random-luokan olio, tarvitaan Update-funktiossa.
        /// </summary>        
        private readonly Random random = new Random();

        /// <summary>
        /// Olion tilanne
        /// </summary>
        /// <remarks>Normaali, Animaatio</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Muodostetaan jokaisesta kirjaimesta oma Label-luokan olio,
        /// joka tallennetaan List-tyypin muuttujaan.
        /// </summary>
        private List<Label> kirjainObjektit = new List<Label>();

        /// <summary>
        /// Apumuuttuja, missä kohtaa kirjaimien paikan animointi on menossa.
        /// </summary>
        private double kirjaimienPaikanVaihe = 0;
                
        /// <summary>
        /// Apumuuttuja, jokaisen kirjaimen animaatioon kulunut aika.
        /// </summary>
        private double[] kuluneetAjat;

        /// <summary>
        /// Apumuuttuja, jokaisen kirjaimen animaation seuraava päivitysaika.
        /// </summary>
        private double[] paivitysAjat;

        /// <summary>
        /// Luokan konstruktori.       
        /// </summary>
        /// <param name="x">Keskipiste X-akselilla.</param>
        /// <param name="y">Keskipiste Y-akselilla.</param>
        /// <param name="kirjasin">Käytettävä kirjasin.</param>
        /// <param name="kirjasimenKoko">Tekstin koko pikseleinä.</param>
        /// <param name="tekstinVari">Tekstin väri.</param>
        /// <param name="teksti">Näytettävä merkkijono.</param>
        public Skrollinaytto(int x, int y, Font kirjasin, int kirjasimenKoko, Color tekstinVari, string teksti) : base(0, 0)
        {
            this.tekstinVari = tekstinVari;

            char[] kirjaimet = teksti.ToCharArray();
            aloituspisteetX = new int[kirjaimet.Length];
            aloituspisteetY = new int[kirjaimet.Length];
            animaationKestot = new double[kirjaimet.Length];
            kuluneetAjat = new double[kirjaimet.Length];
            paivitysAjat = new double[kirjaimet.Length];

            int aloitusPositioX = x - (kirjaimet.Length * kirjasimenKoko / 2) / 2 + kirjasimenKoko / 2;

            // muodostetaan jokaisesta merkkijonon kirjaimesta oma Label-luokan olio,
            // jotta kirjaimien sijaintia voidaan animoida.
            for (int i = 0; i < kirjaimet.Length; i++)
            {
                kirjainObjektit.Add(new Label(kirjasimenKoko, kirjasimenKoko, kirjaimet[i].ToString()));                
                kirjainObjektit.Last().X = aloitusPositioX;
                kirjainObjektit.Last().Y = y;
                kirjainObjektit.Last().Font = kirjasin;
                kirjainObjektit.Last().Font.Size = kirjasimenKoko;
                kirjainObjektit.Last().TextColor = tekstinVari;

                Instance.Add(kirjainObjektit.Last(), (int)OlioidenPiirtotasot.UI);

                aloituspisteetX[i] = aloitusPositioX;
                aloituspisteetY[i] = y;                               

                animaationKestot[i] = random.Next(animaationKestonVaihteluvalit[0], animaationKestonVaihteluvalit[1]);
                kuluneetAjat[i] = random.Next(animaationKestonVaihteluvalit[1]);
                paivitysAjat[i] = 0;

                aloitusPositioX += kirjasimenKoko / 2;
            }
        }


        /// <summary>
        /// Aloitetaan tekstin animointi.
        /// </summary>
        public void AsetaAnimaatio()
        {            
            olionTilanne = OlionTilanne.Animaatio;
        }


        /// <summary>
        /// Tuhotaan kirjaimet ja tämä olio.
        /// </summary>
        public override void Destroy()
        {
            foreach(var item in kirjainObjektit)
            {
                item.Destroy();                
            }
            kirjainObjektit.Clear();

            base.Destroy();
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin ajoin.
        /// Päivitetään kirjainten väri ja paikka.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (IsDestroyed) { }
            else if (olionTilanne == OlionTilanne.Animaatio)
            {
                kirjaimienPaikanVaihe += KIRJAIMIEN_PAIKAN_MUUTOS_PER_PAIVITYS;
                
                for (int i = 0; i < kirjainObjektit.Count(); i++)
                {
                    kuluneetAjat[i] = kuluneetAjat[i] + time.SinceLastUpdate.TotalMilliseconds;
                    paivitysAjat[i] = paivitysAjat[i] + ANIMAATION_PAIVITYSAJAN_MUUTOS;

                    if (kuluneetAjat[i] >= animaationKestot[i])
                    {
                        // palataan alkutilanteeseen
                        kirjainObjektit[i].TextColor = tekstinVari;                        
                        kuluneetAjat[i] = kuluneetAjat[i] * ANIMAATION_KULUNEEN_AJAN_KERROIN; 
                    }
                    else if (kuluneetAjat[i] >= 0 && paivitysAjat[i] >= ANIMAATION_PAIVITYSVALI)
                    {
                        // animoidaan tekstiä, vaihdetaan kirjaimen väri
                        kirjainObjektit[i].TextColor = varit[random.Next(varit.Count())];      
                        paivitysAjat[i] = -ANIMAATION_PAIVITYSVALI;                        
                    }

                    // lasketaan kirjaimelle uusi sijainti
                    kirjainObjektit[i].X = aloituspisteetX[i] + (Math.Cos(kirjaimienPaikanVaihe) * aloituspisteetX[i] * ANIMAATIO_MUUTOS_VAAKASUUNNASSA_KERROIN);
                    kirjainObjektit[i].Y = aloituspisteetY[i] + (Math.Sin(kirjaimienPaikanVaihe + i) * ANIMAATIO_MUUTOS_PYSTYSUUNNASSA_KERROIN);
                }                           
            }

            base.Update(time);
        }       

    }

}
