﻿using Jypeli;
using static Jypeli.Game;
using System;
using System.Collections.Generic;
using System.Linq;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Elämänäytön luokka (esittää pelaajan "elämien" määrän symboleina)
    /// </summary>
    class Symbolinaytto : GameObject
    {
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu tai animaatio on suoritettu loppuun
        /// Lopetusanimaatio = olio suorittaa animaatiota              
        /// </code>
        /// </remarks>     
        private enum OlionTilanne : int { Normaali, Lopetusanimaatio }

        /// <summary>
        /// Lopetusanimaatiossa pallon kokoa pienennetään tämän kertoimen mukaan.
        /// </summary>
        private const float LOPETUSANIMAATIO_KOON_KERROIN = 0.9f;

        /// <summary>
        /// Animaatio lopetetaan kun pallon koko on pienempi.
        /// </summary>
        private const float LOPETUSANIMAATIO_PIENIN_KOON_KERROIN = 0.1f;

        /// <summary>
        /// Varjokuvan siirto pikseleinä vaakasuunnassa symboliin verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_VAAKASUUNNASSA = -3;

        /// <summary>
        /// Varjokuvan siirto pikseleinä pystysuunnassa symboliin verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_PYSTYSUUNNASSA = -3;

        /// <summary>
        /// Miten symbolit tasataan niille varattuun tilaan.
        /// </summary>
        /// <remarks>Vasemmalle, Oikealle, Keskelle</remarks>
        private readonly Tasaus tasaus;

        /// <summary>
        /// Symbolin kuva.
        /// </summary>
        private readonly Image symbolinKuva;

        /// <summary>
        /// Symbolin varjon kuva.
        /// </summary>
        private readonly Image symbolinVarjonKuva;

        /// <summary>
        /// Yhden symbolin leveys pikseleinä.
        /// </summary>        
        private readonly int symbolinLeveys;

        /// <summary>
        /// Yhden symbolin korkeus pikseleinä.
        /// </summary>
        private readonly int symbolinKorkeus;

        /// <summary>
        /// Symbolien väliin jätettävä tyhjä tila pikseleinä.
        /// </summary>
        private readonly int symbolienValinenMarginaali;

        /// <summary>
        /// Symbolien esittämiseen varatun tilan leveys pikseleinä.
        /// </summary>
        private readonly int symbolinaytonLeveys;

        /// <summary>
        /// Symboleille varatun tilan keskipiste x-akselilla.
        /// </summary>
        private readonly int symbolinaytonKeskipisteX;

        /// <summary>
        /// Symboleille varatun tilan keskipiste y-akselilla.
        /// </summary>
        private readonly int symbolinaytonKeskipisteY;

        /// <summary>
        /// Olion tilanne
        /// </summary>
        /// <remarks>Normaali, Lopetusanimaatio</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Jokaisesta symbolista luodaan oma GameObject-luokan olio,
        /// joka tallennetaan List-tyypin muuttujaan.
        /// </summary>
        private List<GameObject> symbolioliot = new List<GameObject>();

        /// <summary>
        /// Jokaisen symbolin varjosta luodaan oma GameObject-luokan olio,
        /// joka tallennetaan List-tyypin muuttujaan.
        /// </summary>
        private List<GameObject> varjoOliot = new List<GameObject>();

        /// <summary>
        /// Tähän listaan siirretään kaikki symboli- ja varjo-oliot jotka poistetaan (animoidaan)
        /// </summary>
        private List<GameObject> poistettavatOliot = new List<GameObject>();

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun animaatio on valmis.</remarks>        
        public event EventHandler SymbolinayttoValmis;

        /// <summary>
        /// Luokan konstruktori.
        /// </summary>
        /// <param name="x">Symboleille varatun tilan keskipiste x-akselilla.</param>
        /// <param name="y">Symboleille varatun tilan keskipiste y-akselilla.</param>
        /// <param name="symbolinaytonLeveys">Symbolien esittämiseen varatun tilan leveys pikseleinä.</param>
        /// <param name="symbolinLeveys">Yhden symbolin leveys pikseleinä.</param>
        /// <param name="symbolinKorkeus">Yhden symbolin korkeus pikseleinä.</param>
        /// <param name="kuva">Symbolin kuva.</param>
        /// <param name="varjokuva">Symbolin varjon kuva.</param>
        /// <param name="marginaali">Symbolien väliin jätettävä tyhjä tila pikseleinä.</param>
        /// <param name="kpl">Kuinka monta symbolia näytetään.</param>
        /// <param name="tasaus">Miten symbolit tasataan niille varattuun tilaan.</param>
        public Symbolinaytto(int x, int y, int symbolinaytonLeveys, int symbolinLeveys, int symbolinKorkeus, Image kuva, Image varjokuva, int marginaali, int kpl, Tasaus tasaus) : base(0, 0)
        {
            if (kpl < 0) kpl = 0;
                        
            symbolinaytonKeskipisteX = x;
            symbolinaytonKeskipisteY = y;
            this.symbolinLeveys = symbolinLeveys;
            this.symbolinKorkeus = symbolinKorkeus;
            symbolinKuva = kuva;
            symbolinVarjonKuva = varjokuva;
            symbolienValinenMarginaali = marginaali;            
            this.tasaus = tasaus;
            this.symbolinaytonLeveys = symbolinaytonLeveys;
        }


        /// <summary>
        /// Päivitetään symbolien määrä.
        /// </summary>
        /// <param name="kpl">Uusi symbolien määrä.</param>
        public void PaivitaSymbolienMaara(int kpl)
        {
            if (kpl < 0) kpl = 0;

            int erotus = kpl - symbolioliot.Count();

            if (erotus == 0) return;

            if (kpl == 0 && symbolioliot.Count() > 0)
            {
                // poistetaan kaikki symbolit
                olionTilanne = OlionTilanne.Lopetusanimaatio;

                foreach (var item in symbolioliot)
                {
                    // lisätään objekti animoitavien listaan
                    poistettavatOliot.Add(item);                    
                }
                symbolioliot.Clear();

                foreach (var item in varjoOliot)
                {
                    // lisätään objekti animoitavien listaan
                    poistettavatOliot.Add(item);
                }
                varjoOliot.Clear();                

                return;
            }

            if (erotus > 0)
            {
                // lisätään symboleita listan loppuun
                for (int i = 0; i < erotus; i++)
                {
                    symbolioliot.Add(new GameObject(symbolinLeveys, symbolinKorkeus));
                    int indeksi = symbolioliot.Count() - 1;
                    symbolioliot[indeksi].Image = symbolinKuva;
                    Instance.Add(symbolioliot[indeksi], (int)OlioidenPiirtotasot.UI);

                    // varjo
                    varjoOliot.Add(new GameObject(symbolinLeveys, symbolinKorkeus));
                    varjoOliot[indeksi].Image = symbolinVarjonKuva;
                    Instance.Add(varjoOliot[indeksi], (int)OlioidenPiirtotasot.UIvarjot);       
                }
            }
            else
            {
                // vähennetään symboleita listan loppupäästä
                olionTilanne = OlionTilanne.Lopetusanimaatio;

                for (; erotus < 0; erotus++)
                {
                    // lisätään objekti animoitavien listaan                    
                    poistettavatOliot.Add(symbolioliot.Last());
                    symbolioliot.RemoveAt(symbolioliot.Count() - 1);                    

                    poistettavatOliot.Add(varjoOliot.Last());
                    varjoOliot.RemoveAt(varjoOliot.Count() - 1);
                }
            }

            // lasketaan jokaiselle symbolille uusi koordinaatti
            int symbolienLeveysYhteensa = kpl * (symbolinLeveys + symbolienValinenMarginaali) - symbolienValinenMarginaali;
            int ensimmainenKoordinaattiX = 0;

            switch (tasaus)
            {
                case Tasaus.Vasemmalle:
                    ensimmainenKoordinaattiX = symbolinaytonKeskipisteX - symbolinaytonLeveys / 2 + symbolinLeveys / 2;
                    break;
                case Tasaus.Oikealle:
                    ensimmainenKoordinaattiX = symbolinaytonKeskipisteX + symbolinaytonLeveys / 2 - symbolienLeveysYhteensa + symbolinLeveys / 2;
                    break;
                default:
                    ensimmainenKoordinaattiX = symbolinaytonKeskipisteX - symbolienLeveysYhteensa / 2 + symbolinLeveys / 2;
                    break;
            }

            for (int i = 0; i < symbolioliot.Count(); i++)
            {
                symbolioliot[i].X = ensimmainenKoordinaattiX + i * (symbolinLeveys + symbolienValinenMarginaali);
                symbolioliot[i].Y = symbolinaytonKeskipisteY;

                varjoOliot[i].X = symbolioliot[i].X + VARJON_SIIRTO_VAAKASUUNNASSA;
                varjoOliot[i].Y = symbolioliot[i].Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
            }

            if (olionTilanne != OlionTilanne.Lopetusanimaatio) SymbolinayttoValmis?.Invoke(this, EventArgs.Empty);
        }


        /// <summary>
        /// Tuhotaan symbolit, varjot ja tämä olio.
        /// </summary>
        public override void Destroy()
        {
            foreach (var item in symbolioliot)
            {
                item.Destroy();
            }
            symbolioliot.Clear();

            foreach (var item in varjoOliot)
            {
                item.Destroy();
            }
            varjoOliot.Clear();

            foreach (var item in poistettavatOliot)
            {
                item.Destroy();
            }
            poistettavatOliot.Clear();

            base.Destroy();
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin ajoin.
        /// Päivitetään symbolin animaatio.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (IsDestroyed) { }
            else if (olionTilanne == OlionTilanne.Lopetusanimaatio)
            {
                if (poistettavatOliot.Count() > 0)
                {
                    // käydään läpi oliot listan lopusta alkuun,
                    // koska jotkut oliot poistetaan ja se muuttaa listan indeksiä
                    for (int i = poistettavatOliot.Count() - 1; i >= 0; i--)
                    {
                        poistettavatOliot[i].Size = poistettavatOliot[i].Size * LOPETUSANIMAATIO_KOON_KERROIN;
                        if (poistettavatOliot[i].Size.X < LOPETUSANIMAATIO_PIENIN_KOON_KERROIN)
                        {
                            poistettavatOliot[i].Destroy();
                            poistettavatOliot.RemoveAt(i);
                            if (poistettavatOliot.Count() == 0) break;
                        }
                    }                    
                }

                if (poistettavatOliot.Count() == 0)
                {
                    olionTilanne = OlionTilanne.Normaali;
                    SymbolinayttoValmis?.Invoke(this, EventArgs.Empty);
                }
            }            

            base.Update(time);
        }

    }

}
