﻿using Jypeli;
using System;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{       
    /// <summary>
    /// Taustakuvan luokka.
    /// </summary>
    /// <remarks>
    /// Taustakuvalla on kaksi erilaista animaatiota, "ympyrän" muotoinen
    /// liike, tai vieritys yhteen suuntaan.
    /// </remarks>
    class Taustakuva : GameObject
    {
        /// <summary>
        /// Jos kuvion (punaisen kanavan) arvo on pienempi kuin raja-arvo,
        /// pikselille annetaan uusi väri.
        /// </summary>
        private const int VARIMUUTOKSEN_RAJA_ARVO = 240;

        /// <summary>
        /// Kuinka paljon taustakuvan sijaintia "ympyrässä" muutetaan per Update-funktion kutsu.
        /// </summary>
        private const double KULMAN_MUUTOS_PER_PAIVITYS = 0.005d;

        /// <summary>
        /// Kuinka nopeasti taustakuvaa vieritetään lineaarisesti.
        /// </summary>
        private const double VIERITYKSEN_NOPEUS = 2.0d;

        /// <summary>
        /// Vierityksen suuntavektori.
        /// </summary>
        private Vector vierityksenSuunta = Vector.Zero;

        /// <summary>
        /// Annetun kuvion leveys pikseleinä.
        /// </summary>
        private int kuvionLeveys;

        /// <summary>
        /// Annetun kuvion korkeus pikseleinä.
        /// </summary>
        private int kuvionKorkeus;

        /// <summary>
        /// Apumuuttuja, käytetään taustakuvan animointiin.
        /// </summary>
        private double kulma = 0;
        
        /// <summary>
        /// Apumuuttuja, taustakuvalle annettava värisävy.
        /// </summary>
        private Color varisavy;
  
        /// <summary>
        /// Luokan konstruktori.
        /// </summary>
        /// <remarks>
        /// Ohjelman ajon aikana luodaan vain yksi taustakuva-luokan olio,
        /// jolle annetaan uusi kuvio aina kun kenttä vaihtuu.
        /// </remarks>
        public Taustakuva() : base(0, 0) {}


        /// <summary>
        /// Muodostetaan uusi taustakuva annetusta kuviosta.
        /// </summary>
        /// <param name="kohdealueenLeveys">Ikkunan leveys pikseleinä.</param>
        /// <param name="kohdealueenKorkeus">Ikkunan korkeus pikseleinä.</param>
        /// <param name="kuvio">Kuvion kuva.</param>
        /// <param name="skaalaus">Kuvion skaalaus (oletusarvo 1.0).</param>
        /// <param name="haluttuVarisavy">Kuviolle annettava värisävy (ei pakollinen).</param>
        public void AsetaTaustakuva(int kohdealueenLeveys, int kohdealueenKorkeus, Image kuvio, float skaalaus = 1.0f, Color? haluttuVarisavy = null)
        {
            // täytetään koko taustakuvan pinta-ala kuviolla            
            if (skaalaus < 0) skaalaus = 0;           

            if (haluttuVarisavy != null && haluttuVarisavy is Color)
            {
                varisavy = (Color)haluttuVarisavy;
            }

            Image = kuvio.Clone();                  
            Image.Rescale((int)(Image.Width * skaalaus), (int)(Image.Height * skaalaus));
            Image.ApplyPixelOperation(Pikselioperaatio);

            kuvionLeveys = Image.Width;
            kuvionKorkeus = Image.Height;

            // taustakuvan leveyden pitää olla vähintään kohdealueen leveys + 2 * kuvion leveys,
            // jotta taustakuvaa voidaan liikuttaa vaakasuunnassa saumattomasti
            //
            // esim. kohdealue on 1000 pikseliä leveä, kuvio on 300 pikseliä leveä,
            // joten kuvioita tarvitaan 1000 / 300 + 3 = 3 + 3 = 6
            // eli taustakuvan leveydeksi tulee 6 * 300 = 1800 pikseliä            
            int kuvionPiirtokerratVaakasuunnassa = kohdealueenLeveys / kuvionLeveys + 3;
            int kuvionPiirtokerratPystysuunnassa = kohdealueenKorkeus / kuvionKorkeus + 3;

            TextureWrapSize = new Vector(kuvionPiirtokerratVaakasuunnassa, kuvionPiirtokerratVaakasuunnassa);

            int taustakuvanLeveys = kuvionPiirtokerratVaakasuunnassa * kuvionLeveys;
            int taustakuvanKorkeus = kuvionPiirtokerratPystysuunnassa * kuvionKorkeus;

            Size = new Vector(taustakuvanLeveys, taustakuvanKorkeus);
        }


        /// <summary>
        /// Kuvio värisävyn muuttamiseen tarvittava apufunktio.
        /// </summary>
        /// <param name="vari">Pikselin väri.</param>
        /// <returns>Pikselin käsitelty väri.</returns>
        private Color Pikselioperaatio(Color vari)
        {
            // oletetaan, että kuvio on mustavalkoinen, siitä luetaan vain punaisen värin komponentti
            if (vari.RedComponent < VARIMUUTOKSEN_RAJA_ARVO) return varisavy;
            return vari;
        }


        /// <summary>
        /// Annetaan taustakuvalle satunnainen suunta.
        /// </summary>
        public void AsetaVierityksenSuuntaSatunnaisesti()
        {          
            vierityksenSuunta = RandomGen.NextVector(1, 1);            
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin ajoin.
        /// Päivitetään taustakuvan positio.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (!IsDestroyed)
            {
                if (vierityksenSuunta == Vector.Zero)
                {
                    kulma += KULMAN_MUUTOS_PER_PAIVITYS;
                    X = (Math.Cos(kulma) * kuvionLeveys) % kuvionLeveys;
                    Y = (Math.Sin(kulma) * kuvionKorkeus) % kuvionKorkeus;
                }
                else
                {
                    X = (X + vierityksenSuunta.X * VIERITYKSEN_NOPEUS) % kuvionLeveys;
                    Y = (Y + vierityksenSuunta.Y * VIERITYKSEN_NOPEUS) % kuvionKorkeus;
                }              
            }
            
            base.Update(time);
        }

    }

}
