﻿using Jypeli;
using System;
using static Jypeli.Game;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Tiiliolion luokka.
    /// </summary>
    class Tiili : PhysicsObject
    {
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu tai animaatio on suoritettu loppuun            
        /// Lopetusanimaatio = olio suorittaa animaatiota, jonka loppumista pitää odottaa            
        /// </code>
        /// </remarks>        
        private enum OlionTilanne : int { Normaali, Lopetusanimaatio }
       
        /// <summary>
        /// Lopetusanimaatiossa tiilen kokoa pienennetään tämän kertoimen mukaan.
        /// </summary>
        private const float LOPETUSANIMAATIO_KOON_KERROIN = 0.9f;

        /// <summary>
        /// Animaatio lopetetaan kun tiilen koko on pienempi.
        /// </summary>
        private const float ANIMAATIO_PIENIN_KOON_KERROIN = 0.1f;

        /// <summary>
        /// Varjokuvan siirto pikseleinä vaakasuunnassa tiileen verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_VAAKASUUNNASSA = -5;

        /// <summary>
        /// Varjokuvan siirto pikseleinä pystysuunnassa tiileen verrattuna.
        /// </summary>
        private const int VARJON_SIIRTO_PYSTYSUUNNASSA = -5;

        /// <summary>
        /// Tiilen yksilöllinen tunniste.
        /// </summary>
        private readonly int id;

        /// <summary>
        /// Tiilen tuhoamisesta saatavat pisteet.
        /// </summary>
        private readonly int pisteet;

        /// <summary>
        /// Olion tilanne.
        /// </summary>
        /// <remarks>Normaali, Lopetusanimaatio</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Varjon GameObject-olio.
        /// </summary>
        private GameObject varjoOlio;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo osuu tiileen.</remarks>
        public event EventHandler<int> LisaaPisteita;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo osuu tiileen.</remarks>
        public event EventHandler<AaniefektienTyypit> SoitaAaniefekti;

        /// <summary>
        /// Tiilitehdas-olio kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pallo osuu tiileen.</remarks>
        public event EventHandler UusiTiiliOnTuhoutumassa;

        /// <summary>
        /// Tiilitehdas-olio kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun tiilen lopetusanimaatio on suoritettu loppuun.</remarks>
        public event EventHandler<int> TiiliTuhoutui;

        /// <summary>
        /// Luokan konstruktori.
        /// </summary>
        /// <param name="id">Yksilöllinen tunniste.</param>
        /// <param name="leveys">Leveys pikseleinä.</param>
        /// <param name="korkeus">Korkeus pikseleinä.</param>
        /// <param name="x">Keskipiste X-akselilla.</param>
        /// <param name="y">Keskipiste Y-akselilla.</param>
        /// <param name="kuva">Tiilen kuva.</param>
        /// <param name="varjonKuva">Tiilen varjon kuva.</param>
        /// <param name="pisteet">Tiilen tuhoamisesta saatavat pisteet.</param>
        public Tiili(int id, int leveys, int korkeus, int x, int y, Image kuva, Image varjonKuva, int pisteet) : base(leveys, korkeus)
        {
            MakeStatic(); // staattinen olio
            Restitution = 1.0; // kimmoisuus, ei muuta pallon nopeutta jos arvo on 1
            KineticFriction = 0;

            this.id = id;
            X = x;
            Y = y;
            Image = kuva;
            this.pisteet = pisteet;

            varjoOlio = new GameObject(Size.X, Size.Y);
            varjoOlio.Image = varjonKuva;
            varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
            varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;
            Instance.Add(varjoOlio, (int)OlioidenPiirtotasot.Varjot);
        }


        /// <summary>
        /// Tätä funktiota kutsutaan kun pallo osuu tiileen.
        /// </summary>
        public void PalloTormasiTiileen()
        {
            IgnoresCollisionResponse = true;
            olionTilanne = OlionTilanne.Lopetusanimaatio;

            // lähetetään viestit
            LisaaPisteita?.Invoke(this, pisteet);
            SoitaAaniefekti?.Invoke(this, AaniefektienTyypit.Tiili);
            UusiTiiliOnTuhoutumassa?.Invoke(this, EventArgs.Empty);            
        }


        /// <summary>
        /// Tuhotaan tiili ja sen varjo.
        /// </summary>
        public override void Destroy()
        {
            varjoOlio.Destroy();
            varjoOlio = null;
            base.Destroy();
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin välein.
        /// Päivittää tiilen animaation.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (IsDestroyed) { }
            else if (olionTilanne == OlionTilanne.Lopetusanimaatio)
            {
                Size = Size * LOPETUSANIMAATIO_KOON_KERROIN;
                varjoOlio.Size = Size;

                varjoOlio.X = X + VARJON_SIIRTO_VAAKASUUNNASSA;
                varjoOlio.Y = Y + VARJON_SIIRTO_PYSTYSUUNNASSA;

                if (Size.X < ANIMAATIO_PIENIN_KOON_KERROIN)
                {
                    // lopetusanimaatio on valmis
                    olionTilanne = OlionTilanne.Normaali;
                    TiiliTuhoutui?.Invoke(this, id);
                }                           
            }

            base.Update(time);
        }

    }

}
