﻿using Jypeli;
using System;
using System.Collections.Generic;
using System.Linq;
using static Jypeli.Game;

/// @author  Micke Saarnela
/// @version 11.10.2023
///
/// <summary>
/// Jyväskylän yliopiston Ohjelmointi 1 -kurssin harjoitustyö, syksy 2023.
/// </summary>

namespace Superpallo3000
{
    /// <summary>
    /// Tiilitehdas-olion luokka. Luo ja poistaa tiilet pelialueelta.
    /// </summary>  
    /// <remarks>
    /// <code>
    /// Käyttö pääohjelmasta:
    /// 1. luodaan uusi tiilitehdasolio
    /// 2. kutsutaan funktiota LisaaTiilet()
    /// 3. odotetaan viestiä TiiletLisatty
    /// 4. lisätään tapahtumankäsittelijät LisaaPisteita += funktio ja SoitaAaniefekti += funktio
    ///    (LisaaPisteita ja SoitaAaniefekti kopioituu jokaiseen tiileen, eli tiilet pitää olla tehty ennen)
    /// </code>
    /// </remarks>
    class Tiilitehdas : GameObject
    {
        /// <summary>   
        /// OlionTilanne tilanne kertoo, missä kohtaa olion koodin suoritus on menossa.
        /// </summary>
        /// 
        /// <remarks>
        /// <code>
        /// Normaali = olio on luotu, tai toiminto on saatu loppuun.
        /// LuodaanTiilet = lisätään tiiliä pelialueelle.
        /// TiiletLuotu = tiilet on luotu.
        /// PoistetaanTiilet = poistetaan jäljellä olevat tiilet pelialueelta.
        /// TiiletPoistettu = kaikki tiilet on poistettu.
        /// </code>
        /// </remarks>
        private enum OlionTilanne : int { Normaali, LuodaanTiilet, TiiletLuotu, PoistetaanTiilet, TiiletPoistettu }
         
        /// <summary>
        /// Yhden tiilen leveys pikseleinä.
        /// </summary>
        private readonly int tiilenLeveys;

        /// <summary>
        /// Yhden tiilen korkeus pikseleinä.
        /// </summary>
        private readonly int tiilenKorkeus;

        /// <summary>
        /// Tiilien väliin jätettävä tyhjä tila pikseleinä.
        /// </summary>
        private readonly int tiilenMarginaali;

        /// <summary>
        /// Tiilen tuhoamisesta saatavat pisteet.
        /// </summary>
        private readonly int pisteetPerTiili;
        
        /// <summary>
        /// Tiilien kuvat.
        /// </summary>
        private readonly Image[] tiilikuvat;

        /// <summary>
        /// Tiilen varjon kuva.
        /// </summary>
        private readonly Image tiilenVarjokuva;

        /// <summary>
        /// Olion tilanne.
        /// </summary>
        /// <remarks>Normaali, LuodaanTiilet, TiiletLuotu, PoistetaanTiilet, TiiletPoistettu</remarks>
        private OlionTilanne olionTilanne = OlionTilanne.Normaali;

        /// <summary>
        /// Tiilien maksimimäärä per kenttä.
        /// </summary>
        private int tiilienMaksimimaara = 1000;

        /// <summary>
        /// Tiilien koordinaatit tallennetaan tähän listaan,
        /// josta valitaan (ja samalla poistetaan) piirrettäviä tiiliä satunnaisesti.
        /// </summary>
        /// <remarks>
        /// taulukko = {tiilen tunnistekoodi, x-koordinaatti, y-koordinaatti}
        /// </remarks>
        private List<int[]> tiilienKoordinaatit = new List<int[]>();

        /// <summary>
        /// Kaikki tiilioliot tallennetaan tähän muuttujaan.
        /// </summary>
        /// <remarks>Avain on tiilen tunniste, Tiili on olio.</remarks>
        private Dictionary<int, Tiili> tiilioliot = new Dictionary<int, Tiili>();

        /// <summary>
        /// Random-luokan olio, tarvitaan useassa funktiossa.
        /// </summary>        
        private readonly Random random = new Random();

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun kaikki tiilet on luotu.</remarks>
        public event EventHandler TiiletLuotu;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun kaikki tiilet on poistettu.</remarks>
        public event EventHandler TiiletPoistettu;

        /// <summary>
        /// Pääohjelma kuuntelee tämän tapahtumankäsittelijän lähettämiä viestejä.
        /// </summary>
        /// <remarks>Viesti lähetetään kun pelaaja on tuhonnut kaikki tiilet, eli kenttä meni läpi.</remarks>
        public event EventHandler KaikkiTiiletTuhottu;

        /// <summary>
        /// Kun pääohjelma lisää Tiilitehdas-oliolle tapahtumankäsittelijän,
        /// se kopioidaan jokaiselle tiilioliolle erikseen.
        /// Tiili lähettää viestin, kun pallo osuu siihen.
        /// </summary>
        public event EventHandler<int> LisaaPisteita
        {
            add 
            {
                foreach (var keyValuePair in tiilioliot)
                {
                    keyValuePair.Value.LisaaPisteita += value;
                }                
            }
            remove
            {
                foreach (var keyValuePair in tiilioliot)
                {
                    keyValuePair.Value.LisaaPisteita -= value;
                }
            }
        }

        /// <summary>
        /// Kun pääohjelma lisää Tiilitehdas-oliolle tapahtumankäsittelijän,
        /// se kopioidaan jokaiselle tiilioliolle erikseen.
        /// Tiili lähettää viestin, kun pallo osuu siihen.
        /// </summary>
        public event EventHandler<AaniefektienTyypit> SoitaAaniefekti
        {
            add
            {
                foreach (var keyValuePair in tiilioliot)
                {
                    keyValuePair.Value.SoitaAaniefekti += value;
                }
            }
            remove
            {
                foreach (var keyValuePair in tiilioliot)
                {
                    keyValuePair.Value.SoitaAaniefekti -= value;
                }
            }
        }
        
        /// <summary>
        /// Luokan konstruktori.
        /// </summary>
        /// <param name="tiilikuvat">Tiilien kuvat taulukkona.</param>
        /// <param name="tiilenVarjokuva">Tiilen varjon kuva.</param>
        /// <param name="tiilenLeveys">Tiilen leveys pikseleinä.</param>
        /// <param name="tiilenKorkeus">Tiilen korkeus pikseleinä.</param>
        /// <param name="tiilenMarginaali">Tiilien väliin jätettävä tyhjä tila pikseleinä.</param>
        /// <param name="pisteetPerTiili">Tiilen tuhoamisesta saatavat pisteet.</param>
        /// <param name="tiilienMaksimimaara">Tiilien enimmäismäärä per kenttä.</param>
        public Tiilitehdas(Image[] tiilikuvat, Image tiilenVarjokuva, int tiilenLeveys, int tiilenKorkeus, int tiilenMarginaali, int pisteetPerTiili, int tiilienMaksimimaara = 1000) : base(0, 0) 
        {
            this.tiilikuvat = tiilikuvat;
            this.tiilenVarjokuva = tiilenVarjokuva;
            this.tiilenLeveys = tiilenLeveys;
            this.tiilenKorkeus = tiilenKorkeus;
            this.tiilenMarginaali = tiilenMarginaali;
            this.pisteetPerTiili = pisteetPerTiili;
            this.tiilienMaksimimaara = tiilienMaksimimaara;
        }


        /// <summary>
        /// Aloitetaan tiilien lisäys pelialueelle.
        /// </summary>
        /// <remarks>
        /// Käytössä yhdeksän erilaista "muodostelmaa",
        /// joista valitaan yksi kentän numeron perusteella.
        /// </remarks>
        /// <param name="kentanNumero">Kentän numero.</param>
        /// <param name="pelialueenLeveys">Tiilille varatun alueen leveys pikseleinä.</param>
        /// <param name="pelialueenKorkeus">Tiilille varatun alueen korkeus pikseleinä.</param>
        /// <param name="ikkunanKorkeus">Ikkunan korkeus pikseleinä.</param>
        /// <param name="ylareunanKorkeus">Pelialueen yläreunan korkeus pikseleinä.</param>
        public void LisaaTiilet(int kentanNumero, int pelialueenLeveys, int pelialueenKorkeus, int ikkunanKorkeus, int ylareunanKorkeus)
        {
            // lasketaan tiilille koordinaatit taulukkoon
            // luodaan vähintään 1 tiili

            tiilienKoordinaatit.Clear();

            int tiilenID = 0;
            int tiiliaVaakasuunnassa = (pelialueenLeveys - tiilenMarginaali * 2) / (tiilenLeveys + tiilenMarginaali);
            int tiiliaPystysuunnassa = pelialueenKorkeus / 2 / (tiilenKorkeus + tiilenMarginaali);

            if (tiiliaVaakasuunnassa <= 0 || tiiliaPystysuunnassa <= 0)
            {
                tiiliaVaakasuunnassa = 1;
                tiiliaPystysuunnassa = 1;
            }

            if (tiilienMaksimimaara < 1) tiilienMaksimimaara = 1;

            if (tiiliaVaakasuunnassa * tiiliaPystysuunnassa > tiilienMaksimimaara)
            {
                tiiliaPystysuunnassa = Math.Max(1, tiilienMaksimimaara / tiiliaVaakasuunnassa);
                tiiliaVaakasuunnassa = Math.Max(1, tiilienMaksimimaara / tiiliaPystysuunnassa);
            }

            int tiilistonLeveys = tiiliaVaakasuunnassa * (tiilenLeveys + tiilenMarginaali) - tiilenMarginaali;
            int tiilistonKorkeus = tiiliaPystysuunnassa * (tiilenKorkeus + tiilenMarginaali) - tiilenMarginaali;
            int pystySiirto = Math.Max(0, ikkunanKorkeus / 2 - ylareunanKorkeus - tiilistonKorkeus);

            for (int y = 0; y < tiiliaPystysuunnassa; y++)
            {
                for (int x = 0; x < tiiliaVaakasuunnassa; x++)
                {
                    // peliobjektin sijainnin x- ja y-arvot tarkoittavat keskipistettä
                    // pelialueen origo sijaitsee (jostain syystä) pelialueen keskipisteessä
                    int posx = -tiilistonLeveys / 2 + x * (tiilenLeveys + tiilenMarginaali) + tiilenLeveys / 2;
                    int posy = pystySiirto + y * (tiilenKorkeus + tiilenMarginaali);
                    tiilenID++;

                    switch (kentanNumero % 9)
                    {
                        // kenttien numerointi alkaa ykkösestä, 9 erilaista kenttää
                        // tässä lasketaan "poistettavat" tiilet, eli ne tiilet,
                        // joita ei lisätä kenttään -> jatketaan silmukkaa (continue)

                        case 1:
                            // joka toinen tiili
                            if (y % 2 == 1 && x % 2 == 0 || y % 2 == 0 && x % 2 == 1) continue;
                            break;

                        case 2:
                            // laatikko
                            if (y > 0 && y < tiiliaPystysuunnassa - 1 && x > 0 && x < tiiliaVaakasuunnassa - 1) continue;
                            break;

                        case 3:
                            // joka toinen rivi
                            if (y % 2 == 1) continue;
                            break;

                        case 4:
                            // laatikko2
                            if (y % 2 == 1 && x > 0 && x < tiiliaVaakasuunnassa - 1) continue;
                            break;

                        case 5:
                            // joka toinen rivi
                            if (y % 2 == 1) continue;
                            break;

                        case 6:
                            // kaksi sisäkkäistä laatikkoa
                            if (tiiliaPystysuunnassa > 6 && tiiliaVaakasuunnassa > 6)
                            {
                                // rivi 0 ja 6 ovat täysiä rivejä
                                if ((y == 1 || y == 5) && x > 0 && x < tiiliaVaakasuunnassa - 1) continue; // rivi 1 & 5
                                if ((y == 2 || y == 4) && (x == 1 || x == tiiliaVaakasuunnassa - 2)) continue; // rivi 2 & 4
                                if (y == 3 && (x == 1 || x == tiiliaVaakasuunnassa - 2 || x > 2 && x < tiiliaVaakasuunnassa - 3)) continue; // rivi 3
                                if (y > 6) continue;
                            }
                            break;

                        case 7:
                            if (x % 2 == 1) continue; // joka toinen sarake
                            break;

                        case 8:
                            if ((x + y) % 3 != 0) continue; // diagonaaliviivoja
                            break;

                        default:
                            // täysi ruudukko
                            break;
                    }

                    tiilienKoordinaatit.Add(new int[] { tiilenID, posx, posy });

                }
            }

            olionTilanne = OlionTilanne.LuodaanTiilet;
        }


        /// <summary>
        /// Lisätään koordinaattitaulukosta yksi satunnainen tiili pelialueelle.
        /// </summary>
        private void LisaaTiili()
        {
            // luodaan yksi tiili ja lisätään se pelialueelle
            // tätä funktiota kutsutaan Update-funktiosta

            if (tiilienKoordinaatit.Count() == 0)
            {
                // kaikki tiilet on jo lisätty, jatketaan muulla ohjelmalla
                olionTilanne = OlionTilanne.TiiletLuotu;
                return;
            }

            // valitaan satunnaiset koordinaatit {tiilenID, x, y}
            int indeksi = random.Next(tiilienKoordinaatit.Count());
            int tiilenID = tiilienKoordinaatit[indeksi][0];
            int posx = tiilienKoordinaatit[indeksi][1];
            int posy = tiilienKoordinaatit[indeksi][2];
            int kuvanIndeksi = random.Next(tiilikuvat.Length);

            // lisätään tiili Dictionaryyn  
            tiilioliot.Add(tiilenID, new Tiili(tiilenID, tiilenLeveys, tiilenKorkeus, posx, posy, tiilikuvat[kuvanIndeksi], tiilenVarjokuva, pisteetPerTiili));
            
            // lisätään tiili pelialueelle
            Instance.Add(tiilioliot[tiilenID], (int)OlioidenPiirtotasot.PeliObjektit);

            // lisätään tapahtumien käsittelyt            
            tiilioliot[tiilenID].TiiliTuhoutui += TiiliTuhoutui;
            tiilioliot[tiilenID].UusiTiiliOnTuhoutumassa += UusiTiiliOnTuhoutumassa;

            // poistetaan käytetty koordinaatti listasta
            tiilienKoordinaatit.RemoveAt(indeksi);
        }


        /// <summary>
        /// Aloitetaan tiilien poistaminen.
        /// </summary>
        public void PoistaTiilet()
        {
            if (tiilioliot.Count() == 0)
            {
                olionTilanne = OlionTilanne.TiiletPoistettu;
                return;
            }

            olionTilanne = OlionTilanne.PoistetaanTiilet;
        }


        /// <summary>
        /// Poistetaan yksi satunnainen tiili pelialueelta.
        /// Kutsutaan vain kuin kaikki tiilet poistetaan.
        /// </summary>
        private void PoistaTiili()
        { 
            if (tiilioliot.Count() == 0)
            {
                olionTilanne = OlionTilanne.TiiletPoistettu;                
                return;
            }

            int[] avaimet = tiilioliot.Keys.ToArray();
            int avain = avaimet[random.Next(avaimet.Count())];

            tiilioliot[avain].Destroy();
            tiilioliot.Remove(avain);
        }


        /// <summary>
        /// Käsitellään tiilioliolta saatu viesti.
        /// Jos viimeinen tiili on tuhottu, lähetetään viesti pääohjelmalle.
        /// </summary>
        /// <param name="sender">Tiiliolio.</param>
        /// <param name="e">Viestin parametrit.</param>
        private void UusiTiiliOnTuhoutumassa(object sender, EventArgs e)
        {           
            if (tiilioliot.Count() <= 1) KaikkiTiiletTuhottu?.Invoke(this, EventArgs.Empty);
        }


        /// <summary>
        /// Käsitellään tiilioliolta saatu viesti.
        /// </summary>
        /// <param name="sender">Tiiliolio.</param>
        /// <param name="id">Tiiliolion tunniste.</param>
        private void TiiliTuhoutui(object sender, int id)
        {           
            Instance.Remove(tiilioliot[id]);
            tiilioliot[id].Destroy();
            tiilioliot.Remove(id);

            if (tiilioliot.Count() == 0)
            {
                // jos kaikki tiilet on tuhottu, lähetetään viesti pääohjelmalle
                TiiletPoistettu?.Invoke(this, EventArgs.Empty);                
            }
        }


        /// <summary>
        /// Jypeli kutsuu tätä funktiota tietyin välein.
        /// Lisätään tai poistetaan tiiliä per päivitys.
        /// </summary>
        /// <param name="time">Aika joka on kulunut viimeisen kutsun jälkeen.</param>
        public override void Update(Time time)
        {
            if (!IsDestroyed)
            {
                switch (olionTilanne)
                {
                    case OlionTilanne.LuodaanTiilet:
                        LisaaTiili();
                        LisaaTiili();
                        break;

                    case OlionTilanne.TiiletLuotu:
                        olionTilanne = OlionTilanne.Normaali;
                        TiiletLuotu?.Invoke(this, EventArgs.Empty);
                        break;

                    case OlionTilanne.PoistetaanTiilet:
                        PoistaTiili();
                        PoistaTiili();
                        break;

                    case OlionTilanne.TiiletPoistettu:
                        olionTilanne = OlionTilanne.Normaali;
                        TiiletPoistettu?.Invoke(this, EventArgs.Empty);
                        break;                   

                    default:
                        break;
                }
            }

            base.Update(time);
        }
        
    }

}
